package com.sillycycle.bagleyd.life3d;

/*
 * @(#)Life3DRules.java
 *
 * Copyright 2008 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The <code>Life3DRules</code> class handles the rules of life.
 * The way implemented should be good for 2 or 3 dimensional life.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life3DRules {
	static final int MAX_NEIGHBORS = 26; // 26 < 32 bits
	static final int MAX_NEIGHBOR_INDEXES = 15;
	static final int BASE = 10;
	static final boolean DEBUG = false;

	private Random generator;
	private String commonName;
	private String discoverer;
	private int neighborIndex;
	private String ruleName;
	private int birth, survival;
	private int birthStir, survivalStir;

	public Life3DRules(String fileName, URL url) {
		readRulesXML(fileName, url, true);
		readRulesXML(fileName, url, false);
		generator = new Random(System.nanoTime());
	}

	static void readRulesXML(String fileName, URL url, boolean pre) {
		try {
			DocumentBuilderFactory docBuilderFactory =
				DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder =
				docBuilderFactory.newDocumentBuilder();
			Document doc;

			if (url == null) {
				doc = docBuilder.parse(new File(fileName));
			} else {
				doc = docBuilder.parse(new URL(url, fileName).openStream());
				//System.out.println("Loaded xml: " + url + fileName);
			}
			int[] ruleIndex = new int[Life3DRules.MAX_NEIGHBOR_INDEXES];
			int ni, ri;
			// normalize text representation
			doc.getDocumentElement().normalize();
			if (DEBUG)
				System.out.println("Root element of the doc is " +
					doc.getDocumentElement().getNodeName());

			NodeList listOfRules =
				doc.getElementsByTagName("rule");
			if (DEBUG)
				System.out.println("Total number of rules: " + listOfRules.getLength());
			if (pre)
				lifeRules = new Life3DRules[Life3DRules.MAX_NEIGHBOR_INDEXES][];
			for (int r = 0; r < listOfRules.getLength(); r++) {
				String neighborRuleString;

				Node ruleNode = listOfRules.item(r);
				if (ruleNode.getNodeType() != Node.ELEMENT_NODE) {
					continue;
				}
				Element ruleElement = (Element) ruleNode;

				NodeList rulenameList = ruleElement.getElementsByTagName("rulename");
				Element rulenameElement = (Element)rulenameList.item(0);

				NodeList textRuleNameList = rulenameElement.getChildNodes();
				neighborRuleString = (textRuleNameList.item(0)).getNodeValue().trim();
				ni = Life3DRules.parseNeighborFromRule(neighborRuleString);
				ri = ruleIndex[ni];
				ruleIndex[ni]++;
				if (pre) {
					continue;
				}
				String commonname = "", discoverer = "";
				NodeList commonnameList = ruleElement.getElementsByTagName("commonname");
				Element commonnameElement = (Element)commonnameList.item(0);

				NodeList textCommonNameList = commonnameElement.getChildNodes();
				if ((textCommonNameList.item(0)) != null)
					commonname = (textCommonNameList.item(0)).getNodeValue().trim();

				NodeList discovererList = ruleElement.getElementsByTagName("discoverer");
				Element discovererElement = (Element)discovererList.item(0);

				NodeList textDiscovererList = discovererElement.getChildNodes();
				if ((textDiscovererList.item(0)) != null)
					discoverer = (textDiscovererList.item(0)).getNodeValue().trim();

				if (DEBUG)
					System.out.println(commonname + " " + discoverer + " " +
						Life3DRules.parseNeighborRule(neighborRuleString));
				lifeRules[ni][ri] = new Life3DRules(commonname, discoverer,
				Life3DRules.parseNeighborRule(neighborRuleString));
			}
			if (pre) {
				for (ni = 0; ni < Life3DRules.MAX_NEIGHBOR_INDEXES; ni++) {
					if (DEBUG)
						System.out.println("ni: " + ni + ", " + ruleIndex[ni]);
					lifeRules[ni] = new Life3DRules[ruleIndex[ni]];
				}
			}
		} catch (SAXParseException err) {
			System.out.println("** Parsing error" + ", line " +
				err.getLineNumber() + ", uri " +
				err.getSystemId());
			System.out.println(" " + err.getMessage());

		} catch (SAXException e) {
			Exception x = e.getException();
			((x == null) ? e : x).printStackTrace();

		} catch (ParserConfigurationException | IOException | DOMException e) {
			e.printStackTrace();
		}
	}

	public Life3DRules(String ruleName, int birth, int survival) {
		this.commonName = "";
		this.discoverer = "";
		this.ruleName = ruleName;
		generator = new Random(System.nanoTime());
		this.birth = birth;
		this.survival = survival;
		this.birthStir = birth;
		this.survivalStir = survival;
	}

	public Life3DRules(String commonName, String discoverer,
			String ruleName) {
		this.commonName = commonName;
		this.discoverer = discoverer;
		this.ruleName = ruleName;
		generator = new Random(System.nanoTime());
		if (ruleName != null && !ruleName.isEmpty()) {
			setMasks(getBirthString(ruleName),
				getSurvivalString(ruleName));
		}
	}

	public Life3DRules(int neighbors, String ruleName) {
		this.commonName = "";
		this.discoverer = "";
		this.neighborIndex = neighborToIndex(neighbors);
		this.ruleName = ruleName;
		generator = new Random(System.nanoTime());
		if (ruleName != null && !ruleName.isEmpty()) {
			setMasks(getBirthString(ruleName),
				getSurvivalString(ruleName));
		}
		init();
	}

	public Life3DRules(int neighbors, int birth, int survival) {
		StringBuilder birthString = new StringBuilder("B");
		StringBuilder survivalString = new StringBuilder("S");
		this.neighborIndex = neighborToIndex(neighbors);
		generator = new Random(System.nanoTime());
		this.birth = birth;
		this.survival = survival;
		this.birthStir = birth;
		this.survivalStir = survival;

		for (int i = 0; i <= MAX_NEIGHBORS; i ++) {
			int mask = 1 << i;
			{
				if ((mask & birth) != 0) {
					if (i >= BASE)
						birthString.append("_");
					birthString.append(i);
				}
				if ((mask & survival) != 0) {
					if (i >= BASE)
						survivalString.append("_");
					survivalString.append(i);
				}
			}
		}
		ruleName = birthString.toString() + "/" + survivalString.toString();
		commonName = "";
		for (Life3DRules lifeRule : lifeRules[neighborIndex]) {
			if (birth == lifeRule.birth && survival == lifeRule.survival) {
				commonName = lifeRule.commonName;
				break;
			}
		}
		discoverer = "";
		for (Life3DRules lifeRule : lifeRules[neighborIndex]) {
			if (birth == lifeRule.birth && survival == lifeRule.survival) {
				discoverer = lifeRule.discoverer;
				break;
			}
		}
		init();
	}

	protected static String toString(int[] values) {
		StringBuilder string = new StringBuilder();

		for (int i = 0; i < values.length; i++) {
			if (i != 0)
				string.append(", ");
			string.append(i + ": " + values[i]);
		}
		return string.toString();
	}

	public String getCommonName() {
		return commonName;
	}

	public String getDiscoverer() {
		return discoverer;
	}

	public int getNeighbors() {
		return indexToNeighbor[neighborIndex];
	}

	public String getRuleName() {
		return ruleName;
	}

	public static String getBirthString(String ruleName) {
		try {
			String[] ruleParts = ruleName.split("/");

			return ruleParts[0].substring(ruleParts[0].indexOf('B') + 1);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public static int getBirth(String ruleName) {
		return setMask(full(getBirthString(ruleName)));
	}

	public static String getSurvivalString(String ruleName) {
		try {
			String[] ruleParts = ruleName.split("/");

			if (ruleParts.length > 1)
				return ruleParts[1].substring(ruleParts[1].indexOf('S') + 1);
			System.out.println("Error" + ruleName);
			return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public static int getSurvival(String ruleName) {
		return setMask(full(getSurvivalString(ruleName)));
	}

	public static String getRuleName(int neighborIndexValue, int ruleIndexValue) {
		return lifeRules[neighborIndexValue][ruleIndexValue].ruleName;
	}

	public int getBirth() {
		return birth;
	}

	public int getSurvival() {
		return survival;
	}

	public void setCommonName(String commonName) {
		this.commonName = commonName;
	}

	public void setDiscoverer(String discoverer) {
		this.discoverer = discoverer;
	}

	public void setNeighborIndex(int neighborIndex) {
		this.neighborIndex = neighborIndex;
	}

	public void setNeighbors(int neighbors) {
		this.neighborIndex = neighborToIndex(neighbors);
	}

	public void setRuleName(String ruleName) {
		this.ruleName = ruleName;
	}

	public void setBirth(int birth) {
		this.birth = birth;
	}

	public void setSurvival(int survival) {
		this.survival = survival;
	}

	public boolean getBirth(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.birth & (1 << count)) != 0);
		}
		return false;
	}

	public boolean getSurvival(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.survival & (1 << count)) != 0);
		}
		return false;
	}

	public boolean getBirthStir(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.birthStir & (1 << count)) != 0);
		}
		return false;
	}

	public boolean getSurvivalStir(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			return ((this.survivalStir & (1 << count)) != 0);
		}
		return false;
	}

	// Use dash for 3 or more consecutive numbers
	public static String condense(String string) {
		StringBuilder sb = new StringBuilder();
		int previous = -2, start = -2;
		int digits = 0;
		boolean gotUnderscore = false;
		boolean putDash = false;

		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);

			if (c == '_') {
				gotUnderscore = true;
				digits = 0;
			} else if (c >= '0' && c <= '9') {
				if (gotUnderscore) {
					digits = (c - '0') * BASE;
					gotUnderscore = false;
				} else {
					int current = digits + (c - '0');
					digits = 0;
					if (current == previous + 1) {
						if (!putDash) {
							if (previous >= BASE) {
								sb.append("_");
							}
							sb.append(previous);
							start = previous;
							putDash = true;
						}
						previous = current;
					} else {
						if (previous >= 0) {
							if (putDash && previous != start + 1) {
								sb.append("-");
							}
							if (previous >= BASE) {
								sb.append("_");
							}
							sb.append(previous);
						}
						previous = current;
						putDash = false;
						start = -2;
					}
				}
			} else {
				if (previous >= 0) {
					if (previous >= BASE) {
						sb.append("_");
					}
					sb.append(previous);
				}
				sb.append(c);
				gotUnderscore = false;
				putDash = false;
				digits = 0;
				previous = -2;
				start = -2;
			}
		}
		if (previous >= 0) {
			if (putDash && previous != start + 1) {
				sb.append("-");
			}
			if (previous >= BASE) {
				sb.append("_");
			}
			sb.append(previous);
		}
		return sb.toString();
	}

	public static String full(String string) {
		StringBuilder sb = new StringBuilder();
		int previous = -2;
		int digits = 0;
		boolean gotUnderscore = false;
		boolean gotDash = false;

		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);

			if (c == '-') {
				gotDash = true;
			} else if (c == '_') {
				gotUnderscore = true;
				digits = 0;
			} else if (c >= '0' && c <= '9') {
				if (gotUnderscore) {
					digits = (c - '0') * BASE;
					gotUnderscore = false;
				} else {
					int current = digits + (c - '0');
					digits = 0;
					if (gotDash) {
						for (int j = previous; j <= current; j++) {
							if (j >= BASE) {
								sb.append("_");
							}
							sb.append(j);
						}
						gotDash = false;
						previous = -2;
					} else {
					if (previous >= 0) {
						if (previous >= BASE) {
							sb.append("_");
							}
							sb.append(previous);
						}
						previous = current;
					}
				}
			} else {
				if (previous >= 0) {
					if (previous >= BASE) {
						sb.append("_");
					}
					sb.append(previous);
				}
				sb.append(c);
				gotUnderscore = false;
				gotDash = false;
				digits = 0;
				previous = -2;
			}
		}
		if (previous >= 0) {
			if (previous >= BASE) {
				sb.append("_");
			}
			sb.append(previous);
		}
		return sb.toString();
	}

	public static int setMask(String string) {
		int mask = 0, j = -1, extraDigits = 0, leadingDigit = 0;

		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);

			// Disable hexadecimal
			//if (c >= 'A' && c <= 'Z') {
			//	j = BASE + c - 'A';
			//	extraDigits = 0;
			//} else
			if (c >= 'a' && c <= 'z') {
				//g = c - 'a';
				if (j != -1) {
					mask -= (1 << j);
					j = -1;
					continue;
				}
			} else if (c >= '0' && c <= '9') {
				if (extraDigits == 0) {
					j = leadingDigit * BASE + (c - '0');
					leadingDigit = 0;
				} else {
					extraDigits--;
					leadingDigit = (c - '0');
					j = -1;
				}
			} else if (c == '_') {
				extraDigits = 1;
				j = -1;
			} else {
				System.out.println("Extra: " + string.charAt(i));
				extraDigits = 0;
				j = -1;
			}
			if (j >= 0 && j <= MAX_NEIGHBORS)
				mask += (1 << j);
		}
		return mask;
	}

	public void setBirthMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.birth |= (1 << count);
		}
		this.birthStir = this.birth;
	}

	public void setSurvivalMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.survival |= (1 << count);
		}
		this.survivalStir = this.survival;
	}

	public void resetBirthMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.birth &= ~(1 << count);
		}
		this.birthStir = this.birth;
	}

	public void resetSurvivalMask(int count) {
		if (count >= 0 && count <= MAX_NEIGHBORS) {
			this.survival &= ~(1 << count);
		}
		this.survivalStir = this.survival;
	}

	// Make sure defaultRuleStrings and defaultRuleIndex are in sync
	static final String[] defaultRuleStrings = {
		"B34/S34", "B34/S34", "B4/S45", "B4/S34",
		"B3/S3", "B4/S34", "B4/S34", "B4/S35",
		"B4/S45", "B4/S46", "B4/S45", "B4/S4",
		"B5/S45", "B5/S6-8", "B6/S78"
	};

	static final Integer defaultRuleIndex(int ni) {
		if (lifeRules[ni] == null)
			return null;
		for (int ri = 0; ri < lifeRules[ni].length; ri++) {
			if (lifeRules[ni][ri] != null &&
					defaultRuleStrings[ni].equals(lifeRules[ni][ri].ruleName))
				return ri;
		}
		return null;
	}

	static Life3DRules[][] lifeRules;

	// Array of Generic Hash Tables
	ArrayList<HashMap<String, Integer>> arrayHashMap = new ArrayList<>();
//http://forums.sun.com/thread.jspa?threadID=530823
	final void init() {
		if (!arrayHashMap.isEmpty())
			arrayHashMap.clear();
		for (int i = 0; i < MAX_NEIGHBOR_INDEXES; i++) {
//System.out.println("Life3DRules.init: " + i);
			arrayHashMap.add(new HashMap<String, Integer>());
//System.out.println("Life3DRules.init: " + neighborIndex + " " + lifeRules[neighborIndex].length);
			for (int j = 0; j < lifeRules[i].length; j++) {
//System.out.println("Life3DRules.init: " + indexToNeighbor[neighborIndex] + Life3DRules.parseRuleFull(lifeRules[i][j].ruleName) + " " + i);
				arrayHashMap.get(i).put(lifeRules[i][j].ruleName, j);
			}
		}
	}

	Integer getRuleIndex(int neighborhoodIndex, String s) {
		if (lifeRules[neighborhoodIndex].length == 0) {
			return null;
		}
//System.out.println("Life3DRules.getRuleIndex: ni " + neighborIndex + " " + lifeRules[neighborIndex].length);
//System.out.println("Life3DRules.getRuleIndex: value " + arrayHashMap.get(neighborIndex).values());
//System.out.println("Life3DRules.getRuleIndex: key " + arrayHashMap.get(neighborIndex).keySet());
//System.out.println("Life3DRules.getRuleIndex: n " + indexToNeighbor[neighborIndex] + " " + s + " " + arrayHashMap.get(neighborIndex).get(s));
		return arrayHashMap.get(neighborhoodIndex).get(s);
	}

	static String[] getStrings(int neighborIndex) {
		String[] strings = new String[lifeRules[neighborIndex].length + 1];
		int i;

		for (i = 0; i < lifeRules[neighborIndex].length; i++) {
			strings[i] = lifeRules[neighborIndex][i].ruleName;
			if (!"".equals(lifeRules[neighborIndex][i].commonName))

				strings[i] += ", " + lifeRules[neighborIndex][i].commonName;
			if (!"".equals(lifeRules[neighborIndex][i].discoverer))
				strings[i] += ", " + lifeRules[neighborIndex][i].discoverer;
		}
		strings[i] = "Other";
		return strings;
		//return (String[]) (arrayHashMap.get(neighborIndex).values()).toArray();
		//Collection<String> strings = arrayHashMap.get(neighborIndex).values();
		//return (String[]) strings.toArray();
		//String[] strings = new String[arrayHashMap.size()];
		//for (int i = 0; i < lifeRules[ni].length; i++) {
	//		strings[i] = arrayHashMap.get(neighborIndex).getKey(i);
		//}
		//return strings;
	}

	static int neighborToPolyhedron(int neighbors) {
		switch (neighbors) {
		case 4: /* Tetrahedron */
		case 16:
		case 22:
			return 4;
		case 6: /* Cube */
		case 8:
		case -14: /* alt 14 */
		case 18:
		case 20:
		case 26:
			return 6;
		/*
			return 8;*/
		case 12: /* Rhombic Dodecahedron, or Edges but every other cube */
		case -18: /* alt 18 */
			return 12;
		case 14: /* Truncated Octahedron, Bitruncated Cubic */
			return 14;
		case 10: /* Octahedron and 2 Tetrahedrons, Alternated Cubic */
		case -26:
		case 38:
			return 16;
		default:
			System.out.println("no polyhedron known for " +
				neighbors + " neighbors");
			return 6;
		}
	}

	static final String[] neighborIndexStrings = {
		"4, Tetrahedron, von Neuman, Faces",
		"6, Cube, von Neuman, Faces",
		"8, Cube, Corners",
		"10, Octahedron Tetrahedron, Faces and Edges (20&10)",

		"12, Rhombic Dodecahedron, Kepler",
		"14, Truncated Octahedron, Kelvin",
		"(-)14, Cube, Faces and Corners",
		"16, Tetrahedrons, no Corners",

		"18, Cube, Faces and Edges",
		"(-)18, Rhombic Dodecahedron, Kepler-Moore",
		"20, Cube, Edges and Corners",
		"22, Tetrahedrons, Gosset-Moore",

		"26, Cube, Moore",
		"(-)26, Octahedron and 2 Tetrahedrons",
		"38, Octahedron and 2 Tetrahedrons (50&38), Gosset-Moore"
	};

	static final int[] indexToPolyhedron = {
		4, 6, 6, 16,  12, 14, 6, 4,  6, 12, 6, 4,  6, 16, 16
	};

	static final int[] indexToNeighbor = {
		4, 6, 8, 10,  12, 14, -14, 16,  18, -18, 20, 22,  26, -26, 38
	};

	static final int[] indexToRealNeighbor = {
		4, 6, 8, 10,  12, 14, 14, 16,  18, 18, 20, 22,  26, 26, 38
	};

	/*static int polyNeighborToIndex(int polyhedron, int neighbor) {
		switch (polyhedron) {
		case 4:
			switch (neighbor) {
			case 4:
				return 0;
			case 16:
				return 7;
			default: // 22
				return 11;
			}
		case 6:
			switch (neighbor) {
			case 6:
				return 1;
			case 8:
				return 2;
			case 12:
				return 4; // Rhombic Dodedahedron
			case -14: // FIXME
			case 14:
				return 6;
			case 18:
				return 8;
			case 20:
				return 10;
			default: // 26
				return 12;
			}
		case 8:
			switch (neighbor) {
			case 6:
				return 2; // Cube
			case 12:
				return 4; // Rhombic Dodecahedron
			default: // 18
				return 9; // Rhombic Dodecahedron
			}
		case 12:
			switch (neighbor) {
			case -18: // FIXME
			case 18:
				return 9;
			default: // 12
				return 4;
			}
		case 14: // 14
			return 5;
		case 16:
			switch (neighbor) {
			case 10:
				return 3;
			case -26: // FIXME
			case 26:
				return 13;
			default: // 38
				return 14;
			}
		default:
			return 12;
		}
	}*/

	static int neighborToIndex(int neighbors) {
		switch (neighbors) {
		case 4: /* Tetrahedron, Faces */
			return 0;
		case 6: /* Cube, Faces */
			return 1;
		case 8: /* Cube, Corners */
			return 2;
		case 10: /* Octahedron and 2 Tetrahedrons, Faces and Edges */
			/* tetra: 4 octa on faces +
				  6 tetra on edges = 10
			   octa: 8 tetra on faces +
				 12 octa on edges = 20 */
			return 3;
		case 12: /* Rhombic Dodecahedron (or octahedron or every other cube Edges) */
			return 4;
		case 14: /* Truncated Octahedron */
			return 5;
		case -14: /* Cube, Faces and Corners */
			return 6;
		case 16: /* Tetrahedrons (corners) */
			/* tetra: 4 * 4 tetra on points = 22 */
			return 7;
		case 18: /* Cube, Faces and Edges */
			return 8;
		case -18: /* Rhombic Doedecahedron (or octahedron), Moore neighborhood */
			return 9;
		case 20: /* Cube, Edges and Corners */
			return 10;
		case 22: /* Tetrahedrons Gosset-Moore */
			/* tetra: 6 tetra on edges +
				  4 * 4 tetra on points = 22 */
			return 11;
		case 26: /* Cube */
			return 12;
		case -26: /* Octahedron and 2 Tetrahedrons, Hokey */
			/* tetra: 4 octa on faces +
				  6 tetra on edges +
				  4 * 4 tetra on points = 26
			   octa: 8 tetra on faces +
				 12 octa on edges +
				 6 octa on points = 26 */
			return 13;
		case 38: /* Octahedron and 2 Tetrahedrons, Gosset-Moore */
			/* tetra: 4 octa on faces +
				  6 tetra on edges +
				  4 * 4 tetra on points +
				  4 * 3 octa on points
				  faces and edges 10, all 38
			   octa: 8 tetra on faces +
				 12 octa on edges +
				 6 octa on points +
				 6 * 4 tetra on points
				 faces and edges 20, all 50 */
			return 14;
		default:
			System.out.println("no neighborhood like " +
				neighbors + " known");
			return Life3DInterface.DEFAULT_NEIGHBOR_INDEX;
		}
	}

	static String parse(String rule) {
		return indexToNeighbor[parseNeighborFromRule(rule)] +
			parseNeighborRule(rule);
	}

	/*static int parsePolyhedronFromRule(String rule) {
		int neighbor = 0;
		int polyhedron = 6;
		int i = 0;
		boolean negative = false;

		for (; i < rule.length(); i++) {
			char letter = rule.charAt(i);

			if (letter >= '0' && letter <= '9')
				neighbor = neighbor * BASE + (letter - '0');
			else if (letter == '-')
				negative = !negative;
			else if (letter == 'N' || letter == 'n')
				// really reading polyhedron not neighbor
				 return neighbor;
			else
				break;
		}
		return polyhedron;
	}*/

	static int parseNeighborFromRule(String rule) {
		int neighbor = 0;
		int i = 0;
		boolean negative = false;

		for (; i < rule.length(); i++) {
			char letter = rule.charAt(i);

			if (letter >= '0' && letter <= '9')
				neighbor = neighbor * BASE + (letter - '0');
			else if (letter == '-')
				negative = !negative;
			else if (letter != 'N')
				// not reading state (not for neighbor)
				break;
		}
		if (neighbor == 0) {
			System.out.println("no neighborhood for rule " +
				rule + " known");
			return Life3DInterface.DEFAULT_NEIGHBOR_INDEX;
		}
		if (negative)
			return neighborToIndex(-neighbor);
//System.out.println("parseNeighborFromRule: " + neighbor);
		return neighborToIndex(neighbor);
	}

	static String parseNeighborRule(String rule) {
		int neighborIndex = parseNeighborFromRule(rule);
		int i = 0, neighbor;

		for (; i < rule.length(); i++) {
			char letter = rule.charAt(i);

			if (!(letter >= '0' && letter <= '9') &&
					letter != '-' &&
					letter != 'N')
				break;
		}
		neighbor = indexToNeighbor[neighborIndex];
		return parseRule(neighbor, rule.substring(i));
	}

	static String parseRule(int neighbors, String rule) {
		int part1, part2;
		StringBuilder[] ruleParts = new StringBuilder[2];
		int i = 0;
		char c;
		int ni = neighborToIndex(neighbors);
		String defaultString = defaultRuleStrings[ni];

		if (DEBUG)
			System.out.println("parseRule: " + rule);
		ruleParts[0] = new StringBuilder();
		ruleParts[1] = new StringBuilder();
		/*Integer defaultInteger = defaultRuleIndex(ni);
		if (defaultInteger != null)
			defaultString = lifeRules[ni][defaultInteger.intValue()].ruleName;
		if (defaultString == null)*/
		if ("".equals(rule)) {
			if (DEBUG)
				System.out.println("parseRule: null default " +
					defaultString);
			return defaultString;
		}
		switch (rule.charAt(i)) {
		// Fertility
		case 'B':
		case 'F':
			part1 = 0;
			break;
		// Environment
		case 'S':
		case 'E':
			part1 = 1;
			break;
		default:
			{
				if (DEBUG)
					System.out.println("parseRule: 1 default " + defaultString);
				return defaultString;
			}
		}
		i++;
		while (i < rule.length()) {
			c = rule.charAt(i);
			if (Character.isDigit(c) || c == '_' || c == '-') {
				ruleParts[part1].append(c);
			} else {
				break;
			}
			i++;
		}
		if (i >= rule.length()) {
			if (DEBUG)
				System.out.println("parseRule: 2 default " + defaultString);
			return defaultString;
		}
		if (rule.charAt(i) == '/') {
			i++;
			if (i >= rule.length()) {
				if (DEBUG)
					System.out.println("parseRule: / default " + defaultString);
				return defaultString;
			}
		}
		switch (rule.charAt(i)) {
		// Fertility
		case 'B':
		case 'F':
			part2 = 0;
			if (part1 == 0) {
				if (DEBUG)
					System.out.println("parseRule: 4 default " + defaultString);
				return defaultString;
			}
			break;
		// Environment
		case 'S':
		case 'E':
			part2 = 1;
			if (part1 == 1) {
				if (DEBUG)
					System.out.println("parseRule: 3 default " + defaultString);
				return defaultString;
			}
			break;
		default:
			{
				if (DEBUG)
					System.out.println("parseRule: 5 default " + defaultString);
				return defaultString;
			}
		}
		i++;
		while (i < rule.length()) {
			c = rule.charAt(i);
			if (Character.isDigit(c) || c == '_' || c == '-') {
				ruleParts[part2].append(c);
			} else {
				break;
			}
			i++;
		}
		if (DEBUG)
			System.out.println("parseRule: " +
				"B" + ruleParts[0].toString() +
				"/S" + ruleParts[1].toString());
		return "B" + ruleParts[0].toString() +
			"/S" + ruleParts[1].toString();
	}

	static String parseRuleFull(String rule) {
		String[] nStrings = rule.split("B");
		String[] sStrings = nStrings[1].split("S");
		String[] bStrings = sStrings[0].split("/");
		String bString = "";
		String sString = "";

		if (sStrings.length > 1)
			sString = sStrings[1];
		if (bStrings.length > 0)
			bString = bStrings[0];
		return nStrings[0] + full("B" + bString) +
			"/" + full("S" + sString);
	}

	static String parseRuleCondense(String rule) {
		String[] nStrings = rule.split("B");
		String[] sStrings = nStrings[1].split("S");
		String[] bStrings = sStrings[0].split("/");
		String bString = "";
		String sString = "";

		if (sStrings.length > 1)
			sString = sStrings[1];
		if (bStrings.length > 0)
			bString = bStrings[0];
		return nStrings[0] + condense("B" + bString) +
				"/" + condense("S" + sString);
	}

	int expand(int number) {
		// find a mask >= 3 where not current in mask but
		// has a "neighbor" that has mask
		int[] ruleMask = new int[Life3DRules.MAX_NEIGHBORS + 1];
		int mask = 1;
		final int lowest = 3;
		int high = lowest;
		int additional = 0;
		int count = 0;

		if (number == 0)
			return 0;
		for (int i = 0; i < Life3DRules.MAX_NEIGHBORS - 1; i++) {
			if ((number & mask) > 0) {
				ruleMask[i] = 1;
				high = i;
			}
			mask <<= 1;
//			System.out.print(ruleMask[i]);
		}
		do {
			additional = generator.nextInt(high + 3 - lowest) + lowest;
//		System.out.println(" " + high + " " + additional + " " + ruleMask[2] + " " + ruleMask[3] + " " + ruleMask[4]);
			count++;
			if (count > 72)
				return number;
			if (additional < 1)
				additional = 1;
			if (additional > Life3DRules.MAX_NEIGHBORS - 1)
				additional = Life3DRules.MAX_NEIGHBORS - 1;
		} while (ruleMask[additional] != 0 ||
			(ruleMask[additional + 1] == 0 &&
			ruleMask[additional - 1] == 0));
//		System.out.println(" " + high + " " + additional + " " + (number | (1 << additional)));
		return number | (1 << additional);
	}

	void setStirs() {
		if (generator.nextInt(2) == 0) {
			birthStir = expand(birth);
			survivalStir = survival;
		} else {
			birthStir = birth;
			survivalStir = expand(survival);
		}
	}

	final void setMasks(String fertility, String environment) {
		birth = setMask(full(fertility));
		survival = setMask(full(environment));
		setStirs();
	}

	// for debugging
	static String toString(int neighbors, Life3DRules param, boolean verbose) {
		int l;
		StringBuilder sb = new StringBuilder("B");

		if (verbose)
			System.out.print("rule (Birth/Survival " +
				neighbors + " neighborhood): ");
		for (l = 0; l <= Math.abs(neighbors) && l < BASE; l++) {
			if (param.getBirth(l)) {
				sb.append(l);
			}
		}
		sb.append("/S");
		for (l = 0; l <= Math.abs(neighbors) && l < BASE; l++) {
			if (param.getSurvival(l)) {
				sb.append(l);
			}
		}
		if (verbose) {
			System.out.println(sb);
			System.out.println("binary rule : Birth 0x" +
				Integer.toHexString(param.birth) +
				", Survival 0x" +
				Integer.toHexString(param.survival));
		}
		return sb.toString();
	}

	public static void main(String[] args) {
		readRulesXML(Life3DInterface.XML_FILE, null, true);
		readRulesXML(Life3DInterface.XML_FILE, null, false);
	}
}
